#%%
import numpy as np
from iminuit import Minuit
from logger_tt import logger
from numba import njit
from scipy.stats import beta as sp_beta
from scipy.stats import betabinom as sp_betabinom
from scipy.stats import expon as sp_exponential

from metaDMG.fit import fit_utils


#%%

priors = fit_utils.get_priors()
q_prior = priors["q"]  # mean = 0.2, concentration = 5
A_prior = priors["A"]  # mean = 0.2, concentration = 5
c_prior = priors["c"]  # mean = 0.1, concentration = 10
phi_prior = priors["phi"]

#%%


@njit
def compute_log_likelihood(A, q, c, phi, x, k, N):
    Dx = A * (1 - q) ** (np.abs(x) - 1) + c
    alpha = Dx * phi
    beta = (1 - Dx) * phi
    return -fit_utils.log_betabinom_PMD(k=k, N=N, alpha=alpha, beta=beta).sum()


@njit
def compute_log_prior(A, q, c, phi):
    lp = (
        fit_utils.log_beta(A, *A_prior)
        + fit_utils.log_beta(q, *q_prior)
        + fit_utils.log_beta(c, *c_prior)
        + fit_utils.log_exponential(phi, *phi_prior)
    )
    return -lp


@njit
def compute_log_posterior(A, q, c, phi, x, k, N):
    log_likelihood = compute_log_likelihood(A=A, q=q, c=c, phi=phi, x=x, k=k, N=N)
    log_p = compute_log_prior(A=A, q=q, c=c, phi=phi)
    return log_likelihood + log_p


#%%


class Frequentist:
    def __init__(
        self,
        data,
        sample,
        tax_id,
        method="posterior",
        p0=None,
        verbose=False,
    ):
        self.sample = sample
        self.tax_id = tax_id
        self.x = data["x"]
        self.k = data["k"]
        self.N = data["N"]
        self.method = method
        self.verbose = verbose
        self._setup_p0(p0)
        self._setup_minuit()
        self.is_fitted = False

    def __repr__(self):
        s = f"Frequentist(data, method={self.method}). \n"
        if self.is_fitted:
            s += self.__str__()
        return s

    def __str__(self):
        if self.is_fitted:
            s = f"sample = {self.sample}, tax_id = {self.tax_id} \n"
            s += f"A = {self.A:.3f}, q = {self.q:.3f},"
            s += f"c = {self.c:.5f}, phi = {self.phi:.1f} \n"
            s += f"damage = {self.damage:.3f} +/- {self.damage_std:.3f} \n"
            s += f"significance = {self.significance:.3f} \n"
            s += f"rho_Ac = {self.rho_Ac:.3f} \n"
            s += f"log_likelihood = {self.log_likelihood:.3f} \n"
            s += f"valid = {self.valid}"
            return s
        else:
            return f"Frequentist(data, method={self.method}, is_fitted=False). \n\n"

    def __call__(self, A, q, c, phi):
        if self.method == "likelihood":
            return self.compute_log_likelihood(
                A=A,
                q=q,
                c=c,
                phi=phi,
            )
        elif self.method == "posterior":
            return self.compute_log_posterior(
                A=A,
                q=q,
                c=c,
                phi=phi,
            )

    def compute_log_likelihood(self, A, q, c, phi):
        return compute_log_likelihood(
            A=A,
            q=q,
            c=c,
            phi=phi,
            x=self.x,
            k=self.k,
            N=self.N,
        )

    def compute_log_posterior(self, A, q, c, phi):
        return compute_log_posterior(
            A=A,
            q=q,
            c=c,
            phi=phi,
            x=self.x,
            k=self.k,
            N=self.N,
        )

    def _setup_p0(self, p0):
        if p0 is None:
            self.p0 = dict(q=0.1, A=0.1, c=0.01, phi=1000)
        else:
            self.p0 = p0

        self.param_grid = {
            "A": sp_beta(*A_prior),
            "q": sp_beta(*q_prior),
            "c": sp_beta(*c_prior),
            "phi": sp_exponential(*phi_prior),
        }

    def _setup_minuit(self, m=None):

        if self.method == "likelihood":
            f = self.compute_log_likelihood

        elif self.method == "posterior":
            f = self.compute_log_posterior

        if m is None:
            self.m = Minuit(f, **self.p0)
        else:
            self.m = m

        if self.method == "likelihood":
            self.m.limits["A"] = (0, 1)
            self.m.limits["q"] = (0, 1)
            self.m.limits["c"] = (0, 1)

        elif self.method == "posterior":
            eps = 1e-10
            self.m.limits["A"] = (0 + eps, 1 - eps)
            self.m.limits["q"] = (0 + eps, 1 - eps)
            self.m.limits["c"] = (0 + eps, 1 - eps)

        self.m.limits["phi"] = (2, None)
        self.m.errordef = Minuit.LIKELIHOOD

    def fit(self):
        if self.verbose:
            print("Initial fit")
        self.m.migrad()
        self.is_fitted = True
        if self.m.valid and self.verbose:
            print("Valid fit")

        # First try to refit it
        if not self.m.valid:
            if self.verbose:
                print("Refitting up to 10 times using last values as p0")
            for i in range(10):
                self.m.migrad()
                if self.m.valid:
                    break
        if self.m.valid and self.verbose:
            print("Valid fit")

        # Then try with a totally flat guess
        if not self.m.valid:
            if self.verbose:
                print("Refitting using a flat p0")
            self._setup_p0({"q": 0.0, "A": 0.0, "c": 0.01, "phi": 100})
            self._setup_minuit()
            self.m.migrad()
        if self.m.valid and self.verbose:
            print("Valid fit")

        # Also try with the default guess
        if not self.m.valid:
            if self.verbose:
                print("Refitting using the default p0")
            self._setup_p0({"q": 0.1, "A": 0.1, "c": 0.01, "phi": 1000})
            self._setup_minuit()
            self.m.migrad()
        if self.m.valid and self.verbose:
            print("Valid fit")

        # If not working, continue with new guesses
        MAX_GUESSES = 100
        if not self.m.valid:

            if self.verbose:
                print("Getting desperate, trying random p0's")

            self.i = 0
            while True:
                p0 = fit_utils.sample_from_param_grid(self.param_grid)
                self._setup_p0(p0)
                self._setup_minuit()
                self.m.migrad()

                if self.m.valid or self.i >= MAX_GUESSES:
                    break
                self.m.migrad()

                if self.m.valid or self.i >= MAX_GUESSES:
                    break

                self.i += 1

        if self.m.valid and self.verbose:
            print(f"Valid fit, number of tries = {self.i}")

        self.valid = self.m.valid

        if self.valid:
            self.damage, self.damage_std, self.significance = self._get_D()
        else:
            self.damage, self.damage_std, self.significance = np.nan, np.nan, np.nan

        return self

    @property
    def log_likelihood(self):
        if self.valid:
            return self.compute_log_likelihood(*self.m.values)
        else:
            return np.nan

    def migrad(self):
        return self.fit()

    def minos(self):
        self.m.minos()
        return self

    @property
    def values(self):
        values = self.m.values.to_dict()
        if self.valid:
            return values
        else:
            for key, val in values.items():
                values[key] = np.nan
            return values

    @property
    def errors(self):
        errors = self.m.errors.to_dict()
        if self.valid:
            return errors
        else:
            for key, val in errors.items():
                errors[key] = np.nan
            return errors

    @property
    def A(self):
        return self.values["A"]

    @property
    def A_std(self):
        return self.errors["A"]

    @property
    def q(self):
        return self.values["q"]

    @property
    def q_std(self):
        return self.errors["q"]

    @property
    def c(self):
        return self.values["c"]

    @property
    def c_std(self):
        return self.errors["c"]

    @property
    def phi(self):
        return self.values["phi"]

    @property
    def phi_std(self):
        return self.errors["phi"]

    def _get_D(self):

        A = self.A
        c = self.c
        phi = self.phi

        N = max(self.N[0], 1)

        mu = A
        std = np.sqrt(A * (1 - A) * (phi + N) / ((phi + 1) * N))
        significance = mu / std
        return mu, std, significance

    @property
    def correlation(self):
        return self.m.covariance.correlation()

    @property
    def rho_Ac(self):
        if self.valid:
            return self.correlation["A", "c"]
        else:
            logger.debug(
                f"Error with: sample = {self.sample}, "
                f"tax_id = {self.tax_id}, "
                f"fit method = {self.method}."
            )
            return np.nan

    @property
    def dist(self):

        A = self.A
        q = self.q
        c = self.c
        phi = self.phi

        x = self.x
        N = self.N

        Dx = A * (1 - q) ** (np.abs(x) - 1) + c

        alpha = Dx * phi
        beta = (1 - Dx) * phi

        dist = sp_betabinom(n=N, a=alpha, b=beta)
        return dist

    @property
    def chi2s(self):
        k = self.k
        dist = self.dist
        chi2s = (dist.mean() - k) ** 2 / dist.std() ** 2
        return chi2s

    @property
    def chi2(self):
        if self.valid:
            return np.sum(self.chi2s)
        else:
            return np.nan


#%%


def make_fits(
    config,
    fit_result,
    data,
    sample,
    tax_id,
    forward_only=None,
    method="posterior",
):
    np.random.seed(42)

    if forward_only is None:
        forward_only = config["forward_only"]

    if forward_only:
        data = {key: val[data["x"] > 0] for key, val in data.items()}

    fit = Frequentist(
        data,
        sample,
        tax_id,
        method=method,
    ).fit()

    vars_to_keep = [
        "damage",
        "damage_std",
        "significance",
        "q",
        "q_std",
        "phi",
        "phi_std",
        "A",
        "A_std",
        "c",
        "c_std",
        "rho_Ac",
        "valid",
    ]

    for var in vars_to_keep:
        fit_result[f"MAP_{var}"] = getattr(fit, var)

    return fit



from pathlib import Path

from ete3 import NCBITaxa
from joblib import Memory


cachedir = ".memoization"
memory = Memory(cachedir, verbose=0)

# @lru_cache
@memory.cache
def extract_descendant_tax_ids(tax, include_subspecies=True):
    """Given either tax_id or tax_name, extract all the descendants' taxIDs.
    Subspecies are automatically included, but can be disables with
    include_subspecies = False.
    """

    ncbi = NCBITaxa()

    # species
    try:
        descendants = set(ncbi.get_descendant_taxa(tax, collapse_subspecies=True))
    except ValueError:
        return []

    # subspecies
    if include_subspecies:
        try:
            descendants |= set(ncbi.get_descendant_taxa(tax, collapse_subspecies=False))
        except ValueError:
            pass

    if Path("taxdump.tar.gz").exists():
        Path("taxdump.tar.gz").unlink()

    return list(descendants)


def include_subspecies(subspecies):
    if len(subspecies) == 1:
        return True
    return False


def apply_tax_id_descendants_filter(d_filter, tax_name, sidebar_left_tax_id_subspecies):
    if tax_name is None:
        return None

    tax_ids = extract_descendant_tax_ids(
        tax_name,
        include_subspecies=include_subspecies(sidebar_left_tax_id_subspecies),
    )
    N_tax_ids = len(tax_ids)
    if N_tax_ids != 0:
        if "tax_id" in d_filter:
            d_filter["tax_ids"].extend(tax_ids)
        else:
            d_filter["tax_ids"] = tax_ids



#%%
import itertools
import warnings
from collections import defaultdict
from math import ceil
from multiprocessing import Pool

import joblib
import numpy as np
import numpyro
import pandas as pd
from logger_tt import logger
from tqdm import tqdm

from metaDMG.errors import BadDataError, FittingError
from metaDMG.fit import bayesian, fit_utils, frequentist
from metaDMG.fit.mismatches import add_reference_count
from metaDMG.utils import Config


numpyro.enable_x64()

#%%

BAYESIAN_MAXIMUM_SIZE = 100

#%%


def get_groupby(df_mismatches):
    return df_mismatches.groupby("tax_id", sort=False, observed=True)


def group_to_numpyro_data(config, group):

    x = np.array(group.iloc[: config["max_position"]]["position"], dtype=int)

    if config["forward_only"]:

        forward = "CT"
        forward_ref = forward[0]

        k = np.array(group.iloc[: config["max_position"]][forward], dtype=int)
        N = np.array(group.iloc[: config["max_position"]][forward_ref], dtype=int)

        data = {"x": x, "k": k, "N": N}
        return data

    else:

        forward = "CT"
        forward_ref = forward[0]
        reverse = "GA"
        reverse_ref = reverse[0]

        k_forward = np.array(group.iloc[: config["max_position"]][forward], dtype=int)
        N_forward = np.array(
            group.iloc[: config["max_position"]][forward_ref], dtype=int
        )

        k_reverse = np.array(group.iloc[-config["max_position"] :][reverse], dtype=int)
        N_reverse = np.array(
            group.iloc[-config["max_position"] :][reverse_ref], dtype=int
        )

        data = {
            "x": np.concatenate([x, -x]),
            "k": np.concatenate([k_forward, k_reverse]),
            "N": np.concatenate([N_forward, N_reverse]),
        }

        return data


#%%


def add_non_CT_GA_mismatches(fit_result, group_in):

    group = group_in.copy()

    for base in ["A", "C", "G", "T"]:
        if base not in group.columns:
            add_reference_count(group, base)

    bases = ["AC", "AG", "AT", "CA", "CG", "GC", "GT", "TA", "TC", "TG"]
    out = []
    for base in bases:
        num = group[base].values
        den = group[base[0]].values
        ratio = np.divide(
            num, den, out=np.zeros(num.shape, dtype=float), where=den != 0
        )
        out.append(ratio)

    # out = np.concatenate(out)
    fit_result["non_CT_GA_damage_frequency_mean"] = np.mean(out, axis=1).mean()
    fit_result["non_CT_GA_damage_frequency_std"] = np.std(out, axis=1).mean()


def add_count_information(fit_result, config, group, data):

    if config["forward_only"]:
        fit_result["N_x=1_forward"] = data["N"][0]
        fit_result["N_x=1_reverse"] = np.nan

        fit_result["N_sum_total"] = data["N"].sum()
        fit_result["N_sum_forward"] = fit_result["N_sum_total"]
        fit_result["N_sum_reverse"] = np.nan

        fit_result["N_min"] = data["N"].min()

        fit_result["k_sum_total"] = data["k"].sum()
        fit_result["k_sum_forward"] = fit_result["k_sum_total"]
        fit_result["k_sum_reverse"] = np.nan

    else:

        fit_result["N_x=1_forward"] = data["N"][0]
        fit_result["N_x=1_reverse"] = data["N"][config["max_position"]]

        fit_result["N_sum_total"] = data["N"].sum()
        fit_result["N_sum_forward"] = data["N"][: config["max_position"]].sum()
        fit_result["N_sum_reverse"] = data["N"][config["max_position"] :].sum()

        fit_result["N_min"] = data["N"].min()

        fit_result["k_sum_total"] = data["k"].sum()
        fit_result["k_sum_forward"] = data["k"][: config["max_position"]].sum()
        fit_result["k_sum_reverse"] = data["k"][config["max_position"] :].sum()

    add_non_CT_GA_mismatches(fit_result, group)


#%%


#%%


def fit_single_group(
    config,
    group,
    mcmm=None,
):

    fit_result = {}

    data = group_to_numpyro_data(config, group)
    sample = config["sample"]
    tax_id = group["tax_id"].iloc[0]

    if data["N"].sum() == 0:
        from metaDMG.fit.serial import _setup_logger

        _setup_logger(config)

        s = f"No data for tax_id {tax_id}, sample = {sample}, i.e. sum(N) = 0. "
        s += "Skipping the fit."
        logger.warning(s)
        return None

    if mcmm is not None:
        try:
            bayesian.make_fits(
                fit_result,
                data,
                mcmm,
            )
        except:
            from metaDMG.fit.serial import _setup_logger

            _setup_logger(config)
            s = f"Error in bayesian.make_fits for tax_id {tax_id}, sample = {sample}."
            s += "Skipping the Bayesian fit."
            logger.warning(s)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")

        frequentist.make_fits(
            config,
            fit_result,
            data,
            sample,
            tax_id,
        )  # fit

    add_count_information(fit_result, config, group, data)

    return fit_result


def compute_fits_seriel(config, df_mismatches, with_progressbar=False):

    # Do not initialise MCMC if config["bayesian"] is False
    mcmm = bayesian.init_mcmc(config)

    groupby = get_groupby(df_mismatches)

    if with_progressbar:
        groupby = tqdm(groupby, total=len(groupby))

    d_fit_results = {}
    for tax_id, group in groupby:
        # break

        if with_progressbar:
            groupby.set_description(f"Fitting Tax ID {tax_id}")

        res = fit_single_group(
            config,
            group,
            mcmm,
            # mcmc_null,
        )

        if res is not None:
            d_fit_results[tax_id] = res

    return d_fit_results


def compute_fits_parallel_worker(df_mismatches_config):
    df_mismatches, config, with_progressbar = df_mismatches_config
    return compute_fits_seriel(
        config=config,
        df_mismatches=df_mismatches,
        with_progressbar=with_progressbar,
    )


def grouper(iterable, n):
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


def use_progressbar(config, position):
    if config["bayesian"]:
        return False

    if config["parallel_samples"] == 1 or len(config["samples"]) == 1:
        if position == 0:
            return True

    return False


def get_list_of_groups(
    config,
    df_mismatches_unique,
    N_splits=None,
    N_maximum_group_size=BAYESIAN_MAXIMUM_SIZE,
):

    tax_ids = df_mismatches_unique["tax_id"].unique()

    if N_splits is None:
        cores_per_sample = config["cores_per_sample"]
        N_splits = cores_per_sample * ceil(
            len(tax_ids) / cores_per_sample / N_maximum_group_size
        )

    tax_id_list = np.array_split(tax_ids, N_splits)

    dfs = []
    for position, tax_ids in enumerate(tax_id_list):
        dfs.append(
            (
                df_mismatches_unique.query(f"tax_id in {list(tax_ids)}"),
                config,
                use_progressbar(config, position),
            )
        )
    return dfs


def compute_fits_parallel(
    config,
    df_mismatches_unique,
):

    cores_per_sample = config["cores_per_sample"]

    dfs = get_list_of_groups(
        config,
        df_mismatches_unique,
        N_splits=cores_per_sample,
    )

    d_fit_results = {}
    with Pool(processes=cores_per_sample) as pool:
        for d_fit_results_ in pool.imap_unordered(
            compute_fits_parallel_worker,
            dfs,
        ):
            d_fit_results.update(d_fit_results_)

    return d_fit_results


def compute_fits_parallel_Bayesian(
    config,
    df_mismatches_unique,
    N_maximum_group_size=BAYESIAN_MAXIMUM_SIZE,
):

    cores_per_sample = config["cores_per_sample"]

    dfs = get_list_of_groups(
        config=config,
        df_mismatches_unique=df_mismatches_unique,
        N_maximum_group_size=N_maximum_group_size,
    )

    do_progressbar = config["parallel_samples"] == 1 or len(config["samples"]) == 1

    if len(dfs) == 1:
        logger.debug(f"Computing Bayesian fits in serial (using 1 core).")
        d_fit_results = compute_fits_seriel(
            config,
            df_mismatches_unique,
            with_progressbar=do_progressbar,
        )
        return d_fit_results

    logger.debug(f"Computing Bayesian fits")

    d_fit_results = {}

    N_chunks = int(np.ceil(len(dfs) / cores_per_sample))

    if N_chunks == 1:

        if do_progressbar:
            dfs[0] = (dfs[0][0], dfs[0][1], True)

        with Pool(processes=cores_per_sample) as pool:
            for d_fit_results_ in pool.imap_unordered(
                compute_fits_parallel_worker,
                dfs,
            ):
                d_fit_results.update(d_fit_results_)
        return d_fit_results

    it = grouper(dfs, cores_per_sample)
    if do_progressbar:
        it = tqdm(it, total=N_chunks, unit="chunks")

    for dfs_ in it:
        # break

        if do_progressbar:
            size = dfs_[0][0]["tax_id"].nunique()
            s = f"Fitting in {N_chunks} chunks, each of size {size}, using {cores_per_sample} core(s)"
            it.set_description(s)

        with Pool(processes=cores_per_sample) as pool:
            for d_fit_results_ in pool.imap_unordered(
                compute_fits_parallel_worker,
                dfs_,
            ):
                d_fit_results.update(d_fit_results_)

    return d_fit_results


#%%


def match_tax_id_order_in_df_fit_results(df_fit_results, df_mismatches):
    tax_ids_all = pd.unique(df_mismatches["tax_id"])
    ordered = [tax_id for tax_id in tax_ids_all if tax_id in df_fit_results.index]
    return df_fit_results.loc[ordered]


def make_df_fit_results_from_fit_results(config, d_fit_results, df_mismatches):
    df_fit_results = pd.DataFrame.from_dict(d_fit_results, orient="index")
    df_fit_results["tax_id"] = df_fit_results.index
    # move_column_inplace(df_fit_results, "tax_id", pos=0)

    df_fit_results = match_tax_id_order_in_df_fit_results(df_fit_results, df_mismatches)
    df_fit_results["sample"] = config["sample"]
    categories = []
    df_fit_results = fit_utils.downcast_dataframe(
        df_fit_results, categories, fully_automatic=False
    )

    df_fit_results = df_fit_results.reset_index(drop=True)

    return df_fit_results


#%%


def compute_duplicates(df_mismatches):

    groupby = get_groupby(df_mismatches)

    duplicate_entries = defaultdict(list)
    for group in groupby:
        key = joblib.hash(group[1][fit_utils.ref_obs_bases].values)
        duplicate_entries[key].append(group[0])
    duplicate_entries = dict(duplicate_entries)

    unique = [tax_ids[0] for tax_ids in duplicate_entries.values()]
    duplicates = {tax_ids[0]: tax_ids[1:] for tax_ids in duplicate_entries.values()}

    return unique, duplicates


def de_duplicate_fit_results(d_fit_results, duplicates):

    for tax_id_unique, tax_ids_non_unique in duplicates.items():

        if not tax_id_unique in d_fit_results:
            logger.warning(f"Could not de-duplicate tax ID {tax_id_unique}.")
            continue

        for tax_id_non_unique in tax_ids_non_unique:
            d_fit_results[tax_id_non_unique] = d_fit_results[tax_id_unique]


def split(strng, sep, pos):
    strng = strng.split(sep)
    return sep.join(strng[:pos]), sep.join(strng[pos:])


#%%


def read_stats_lca(config: Config):

    import gzip
    from io import StringIO

    columns = [
        "tax_id",
        "tax_name",
        "tax_rank",
        "N_alignments",
        "N_reads",
        "mean_L",
        "var_L",
        "mean_GC",
        "var_GC",
        "tax_path",
    ]

    # TODO: remove when Thorfinn updates his code
    with gzip.open(config["path_mismatches_stat"], "rt") as f:
        data = f.read()

    data = data.replace("'", "")  # remove '
    data = data.replace("\t\t", "\t'")
    data = data.replace('1:root:"no rank"', """1:root:"no rank"'""")

    string_columns = ["tax_id", "tax_name", "tax_rank", "tax_path"]

    df_stats = pd.read_csv(
        StringIO(data),  # config["path_mismatches_stat"]
        sep="\t",
        skiprows=1,
        index_col=False,
        quotechar="'",
        names=columns,
        dtype={column: "string" for column in string_columns},
    )

    for column in ("tax_name", "tax_rank"):
        df_stats[column] = df_stats[column].str.replace('"', "")

    df_stats["std_L"] = np.sqrt(df_stats["var_L"])
    df_stats["std_GC"] = np.sqrt(df_stats["var_GC"])
    return df_stats


def read_stats_non_lca(config: Config):

    columns = ["tax_id", "N_reads", "mean_L", "var_L", "mean_GC", "var_GC"]

    df_stats = pd.read_csv(
        config["path_mismatches_stat"],
        sep="\t",
        names=columns,
        usecols=columns,
    )

    df_stats["std_L"] = np.sqrt(df_stats["var_L"])
    df_stats["std_GC"] = np.sqrt(df_stats["var_GC"])
    return df_stats


def read_stats(config: Config):
    if config["damage_mode"] == "lca":
        return read_stats_lca(config)
    else:
        return read_stats_non_lca(config)


def cut_minimum_reads(
    config: Config,
    df_stats: pd.DataFrame,
) -> pd.DataFrame:
    return df_stats.query(f"(N_reads >= {config['min_reads']})")


def filter_tax_ids(
    config: Config,
    df_stats: pd.DataFrame,
    df_stat_cut: pd.DataFrame,
    df_mismatches: pd.DataFrame,
) -> pd.DataFrame:
    all_tax_ids = set(df_stats["tax_id"].unique())
    good_tax_ids = set(df_stat_cut["tax_id"].unique())
    bad_tax_ids = all_tax_ids.difference(good_tax_ids)
    if len(bad_tax_ids) > 0:
        logger.debug(f"Dropping the following Tax IDs due to too few reads:")
        logger.debug(bad_tax_ids)
        logger.debug(f"with a minimum reads threshold of {config['min_reads']}.")
    return df_mismatches.query("tax_id in @good_tax_ids")


def filter_k_sum(
    config: Config,
    df_mismatches: pd.DataFrame,
) -> pd.DataFrame:

    # filter out tax_id's with 0 k_sum_total
    tax_ids_to_drop = set(df_mismatches.query("k_sum_total == 0")["tax_id"].unique())
    if len(tax_ids_to_drop) > 0:
        logger.debug(f"Dropping the following Tax IDs since k_sum_total == 0:")
        logger.debug(tax_ids_to_drop)
    return df_mismatches.query("k_sum_total > 0")


def filter_max_N_in_group(
    config: Config,
    df_mismatches: pd.DataFrame,
) -> pd.DataFrame:

    # filter out tax_id's with 0 k_sum_total
    tax_ids_to_drop = set(df_mismatches.query("max_N_in_group == 0")["tax_id"].unique())
    if len(tax_ids_to_drop) > 0:
        logger.debug(f"Dropping the following Tax IDs since max_N_in_group == 0:")
        logger.debug(tax_ids_to_drop)
    return df_mismatches.query("max_N_in_group > 0")


#%%


def compute(config, df_mismatches):

    df_stats = read_stats(config)
    df_stat_cut = cut_minimum_reads(config, df_stats)

    # filter out tax_id's with too few reads
    df_mismatches = filter_tax_ids(config, df_stats, df_stat_cut, df_mismatches)

    if len(df_mismatches) == 0:
        s = f"{config['sample']} cut on min reads > {config['min_reads']} made data empty."
        logger.debug("WARNING: BadDataError")
        logger.debug(s)
        raise BadDataError(s)

    # filter out tax_id's with no data in them (max_N_in_group == 0)
    df_mismatches = filter_max_N_in_group(config, df_mismatches)

    if len(df_mismatches) == 0:
        s = f"{config['sample']} df_mismatches.query('max_N_in_group > 0') is empty."
        logger.debug("WARNING: BadDataError")
        logger.debug(s)
        raise BadDataError(s)

    # # filter out tax_id's with 0 k_sum_total
    # df_mismatches = filter_k_sum(config, df_mismatches)

    # if len(df_mismatches) == 0:
    #     s = f"{config['sample']} df_mismatches.query('k_sum_total > 0') is empty."
    #     logger.debug("WARNING: BadDataError")
    #     logger.debug(s)
    #     raise BadDataError(s)

    # find unique tax_ids (when compairing the mismatches matrices)
    # such that only those are fitted
    unique, duplicates = compute_duplicates(df_mismatches)

    logger.debug(
        f"Instead of fitting all {df_mismatches['tax_id'].nunique()} tax IDs, "
        f"only fit the {len(unique)} unique ones."
    )

    df_mismatches_unique = df_mismatches.query(f"tax_id in {unique}")

    if config["bayesian"]:
        # logger.debug(f"Computing Bayesian fits")
        d_fit_results = compute_fits_parallel_Bayesian(config, df_mismatches_unique)

    else:
        if config["cores_per_sample"] == 1:
            logger.debug(f"Fitting in seriel.")

            if config["parallel_samples"] == 1 or len(config["samples"]) == 1:
                with_progressbar = True
            else:
                with_progressbar = False

            d_fit_results = compute_fits_seriel(
                config,
                df_mismatches_unique,
                with_progressbar=with_progressbar,
            )

        else:
            s = f"Fitting in parallel with {config['cores_per_sample']} cores."
            logger.debug(s)
            d_fit_results = compute_fits_parallel(
                config,
                df_mismatches_unique,
            )

    de_duplicate_fit_results(d_fit_results, duplicates)

    df_fit_results = make_df_fit_results_from_fit_results(
        config,
        d_fit_results,
        df_mismatches,
    )

    df_fit_results = pd.merge(df_fit_results, df_stat_cut, on="tax_id")

    prefix = "" if config["bayesian"] else "MAP_"
    cols_ordered = [
        "sample",
        "tax_id",
        "tax_name",
        "tax_rank",
        "N_reads",
        "N_alignments",
        f"{prefix}damage",
        f"{prefix}significance",
        "mean_L",
        "std_L",
        "mean_GC",
        "std_GC",
        "tax_path",
    ]
    cols_ordered = cols_ordered + [
        c for c in df_fit_results.columns if c not in cols_ordered
    ]

    # if local or global damage
    if config["damage_mode"] in ("local", "global"):
        for col in ["tax_name", "tax_rank", "N_alignments", "tax_path"]:
            cols_ordered.remove(col)

    df_fit_results = df_fit_results[cols_ordered]

    return df_fit_results


# %%
#



#%%

from enum import Enum
from functools import partial
from pathlib import Path
from typing import Iterable, Optional, Union

import click
import typer
import yaml
from rich import print


#%%


output_dir_default = Path("./data/")
# results_dir_default = output_dir_default / "results"
config_file_default = Path("config.yaml")


#%%


# class OrderedCommands(Group):
class OrderedCommands(typer.core.TyperGroup):
    def list_commands(self, ctx: click.Context) -> Iterable[str]:
        return self.commands.keys()


def get_cli_app():
    cli_app = typer.Typer(cls=OrderedCommands, rich_markup_mode="rich")
    return cli_app


def version_callback(value: bool):
    from metaDMG.__version__ import __version__

    if value:
        print(f"metaDMG CLI, version: {__version__}")
        raise typer.Exit()


def is_in_range_or_None(
    x: Optional[float], val_min: float, val_max: float
) -> Optional[float]:
    """Confirms that x is val_min <= x <= val_max

    Parameters
    ----------
    x
        Value to check
    val_min
        Minimum
    val_max
        Maximum

    Returns
    -------
        Confirmed value

    Raises
    ------
    typer.BadParameter
        If x is outside bounds
    """

    if x is None:
        return None

    if x < val_min or val_max < x:
        raise typer.BadParameter(
            f"x has to be between {val_min} and {val_max}. Got: {x}"
        )
    return x


def is_positive_int_or_None(x: Optional[int]) -> Optional[int]:
    """Confirms that x is 0 <= x

    Parameters
    ----------
    x
        Value to check

    Returns
    -------
        Confirmed value

    Raises
    ------
    typer.BadParameter
        If x is outside bounds
    """

    if x is None:
        return None

    if x < 0:
        raise typer.BadParameter(f"x has to be positive. Got: {x}")

    return x


def path_exists_or_None(x: Union[str, Path, None]) -> Union[str, Path, None]:
    """Confirms that x exists or is None

    Parameters
    ----------
    x
        Value to check

    Returns
    -------
        Confirmed value

    Raises
    ------
    typer.BadParameter
        If x is outside bounds
    """

    if x is None:
        return None

    if isinstance(x, str):
        path = Path(x)
    else:
        path = x

    if not path.is_file():
        raise typer.BadParameter(f"The file {path} does not exists.")

    return x


#%%
class RANKS(str, Enum):
    "Ranks allowed in the LCA"

    family = "family"
    genus = "genus"
    species = "species"
    none = ""

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def str_list(cls):
        return [c if c != "" else "none" for c in cls.list()]


class DAMAGE_MODE(str, Enum):
    "Damage mode allowed in the LCA"

    LCA = "lca"
    LOCAL = "local"
    GLOBAL = "global"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def upper_list(cls):
        return [c.upper() for c in cls.list()]


#%%


def set_min_max_similarity_score_edit_dist(
    min_similarity_score: Optional[float],
    max_similarity_score: Optional[float],
    min_edit_dist: Optional[int],
    max_edit_dist: Optional[int],
) -> dict[str, float]:

    if any([min_similarity_score, max_similarity_score]) and any(
        [min_edit_dist, max_edit_dist]
    ):
        raise typer.BadParameter(
            f"You cannot use both similarity scores and edit distances at the same time."
        )

    # edit distances
    if any([min_edit_dist, max_edit_dist]):

        if all([min_edit_dist, max_edit_dist]):

            if min_edit_dist > max_edit_dist:
                raise typer.BadParameter(
                    f"min-edit-dist ({min_edit_dist}) "
                    f"has to be lower than max-edit-dist ({max_edit_dist})"
                )

            return {
                "min_edit_dist": min_edit_dist,
                "max_edit_dist": max_edit_dist,
            }

        else:
            raise typer.BadParameter(
                f"If using (absolute) edit distances, you have to set "
                "both `min_edit_dist` and `max_edit_dist`."
            )

    # similarity scores

    if min_similarity_score is None:
        min_similarity_score = 0.95

    if max_similarity_score is None:
        max_similarity_score = 1.0

    if min_similarity_score > max_similarity_score:
        raise typer.BadParameter(
            f"min-similarity-score ({min_similarity_score}) "
            f"has to be lower than max-similarity-score ({max_similarity_score})"
        )

    return {
        "min_similarity_score": min_similarity_score,
        "max_similarity_score": max_similarity_score,
    }


#%%


def path_endswith(path: Path, s: str) -> bool:
    return str(path.name).endswith(s)


def extract_name(
    filename: Path,
    max_length: int = 100,
    prefix: str = "",
    suffix: str = "",
    long_name: bool = False,
) -> str:
    """Extract the name from a file

    Parameters
    ----------
    filename
        The input file
    max_length
        The maximum length of the name, by default 100
    prefix
        The prefix to be added to the name, by default ""
    suffix
        The suffix to be added to the name, by default ""
    long_name
        Whether or not to use the full name, by default False

    Returns
    -------
        The name
    """
    name = Path(filename).stem
    if not long_name:
        name = name.split(".")[0]
    if len(name) > max_length:
        name = name[:max_length] + "..."
    name = prefix + name + suffix
    return name


def extract_names(file_list, **kwargs):
    return list(map(partial(extract_name, **kwargs), file_list))


def extract_alignment_files(paths: list[Path]) -> list[Path]:
    """Extract all alignment files from a list of paths.
    Alignment files are expected to be .bam, .sam, or .sam.gz.

    Parameters
    ----------
    paths
        Input list of paths

    Returns
    -------
        Output list of alignment files
    """

    from natsort import os_sorted

    alignments = []
    suffixes = (".bam", ".sam", ".sam.gz")

    for path in os_sorted(paths):
        # break
        if path.is_file() and any(path_endswith(path, suffix) for suffix in suffixes):
            alignments.append(path)

        elif path.is_dir():

            files = [
                p
                for p in Path(path).glob("*")
                if any(path_endswith(p, suffix) for suffix in suffixes)
            ]

            recursive = extract_alignment_files(files)
            alignments.extend(recursive)

    return alignments


def extract_samples(
    paths: list[Path],
    prefix: str = "",
    suffix: str = "",
    long_name: bool = False,
) -> dict:
    """Extract all alignment files from a list of files.
    Truncates the name of the files, controlled by prefix, suffix, and long_name

    Parameters
    ----------
    paths
        List of paths to be extracted
    prefix
        The prefix to be added to the name, by default ""
    suffix
        The suffix to be added to the name, by default ""
    long_name
        Whether or not to use the full name, by default False

    Returns
    -------
        Dictionary with names as keys and files as values.
    """

    alignments = extract_alignment_files(paths)
    samples = extract_names(
        alignments,
        prefix=prefix,
        suffix=suffix,
        long_name=long_name,
    )

    if len(alignments) == 0:
        raise typer.BadParameter("No alignment files found.")

    d_alignments = {}
    for sample, path in zip(samples, alignments):
        d_alignments[sample] = str(path)

    return d_alignments


def paths_to_strings(
    d: dict,
    ignore_keys: Optional[Iterable] = None,
) -> dict:
    """Convert all the paths in a dictionary to strings

    Parameters
    ----------
    d
        Input dict to be converted
    ignore_keys
        Ignore the following keys in the iterable, by default None

    Returns
    -------
        Dictionary with strings instead of paths
    """

    if ignore_keys is None:
        ignore_keys = []

    d_out = {}
    for key, val in d.items():
        if val in ignore_keys:
            continue
        elif isinstance(val, list):
            d_out[key] = list(map(str, val))
        elif isinstance(val, tuple):
            d_out[key] = tuple(map(str, val))
        elif isinstance(val, dict):
            d_out[key] = paths_to_strings(val)
        elif isinstance(val, Path):
            d_out[key] = str(val)
        else:
            d_out[key] = val
    return d_out


def save_config_file(
    config: dict,
    config_file: Path,
    overwrite_config: bool = False,
    verbose: bool = True,
) -> None:
    """Save the config file.
    Does not overwrite if file already exists, unless explicitly specified.

    Parameters
    ----------
    config
        Input dict
    config_file
        Save location

    Raises
    ------
    typer.Abort
        Do not overwrite automatically
    """

    if not ("yaml" in config_file.name or "yml" in config_file.name):
        config_file = config_file.with_suffix(".yaml")

    if not overwrite_config:
        if config_file.is_file():
            s = "Config file already exists. Do you want to overwrite it?"
            confirmed = typer.confirm(s)
            if not confirmed:
                print("Exiting")
                raise typer.Abort()

    with open(config_file, "w") as file:
        yaml.dump(config, file, sort_keys=False)
    if verbose:
        print(f"{str(config_file)} was created")


#%%


def get_config_dict(
    samples: list[Path],
    names: Optional[Path] = None,
    nodes: Optional[Path] = None,
    acc2tax: Optional[Path] = None,
    min_similarity_score: Optional[float] = None,
    max_similarity_score: Optional[float] = None,
    min_edit_dist: Optional[int] = None,
    max_edit_dist: Optional[int] = None,
    min_mapping_quality: int = 0,
    custom_database: bool = False,
    lca_rank: RANKS = RANKS.none,
    metaDMG_cpp: str = "./metaDMG-cpp",
    max_position: int = 15,
    min_reads: int = 0,
    weight_type: int = 1,
    forward_only: bool = False,
    bayesian: bool = False,
    output_dir: Path = output_dir_default,
    parallel_samples: int = 1,
    cores_per_sample: int = 1,
    config_file: Path = config_file_default,
    sample_prefix: str = "",
    sample_suffix: str = "",
    long_name: bool = False,
    damage_mode: DAMAGE_MODE = DAMAGE_MODE.LCA,
    __version__: str = "",
) -> dict:

    config = paths_to_strings(
        {
            "samples": extract_samples(
                samples,
                prefix=sample_prefix,
                suffix=sample_suffix,
                long_name=long_name,
            ),
            #
            "metaDMG_cpp": metaDMG_cpp,
            "names": names,
            "nodes": nodes,
            "acc2tax": acc2tax,
            **set_min_max_similarity_score_edit_dist(
                min_similarity_score, max_similarity_score, min_edit_dist, max_edit_dist
            ),
            "min_mapping_quality": min_mapping_quality,
            "lca_rank": lca_rank.lower(),
            "max_position": max_position,
            "min_reads": min_reads,
            "weight_type": weight_type,
            "custom_database": custom_database,
            "forward_only": forward_only,
            #
            "output_dir": output_dir,
            "parallel_samples": parallel_samples,
            "cores_per_sample": cores_per_sample,
            "bayesian": bayesian,
            "config_file": str(config_file),
            "damage_mode": damage_mode.lower(),
            "version": __version__,
        }
    )

    return config



#%%

import numpy as np
import pandas as pd

from metaDMG import __version__ as version
from metaDMG.fit import mismatches


#%%


def make_reverse_group(group):
    group_reverse = pd.DataFrame(0, index=group.index, columns=group.columns)
    group_reverse.loc[:, "tax_id"] = group.loc[:, "tax_id"]
    group_reverse.loc[:, "direction"] = "3'"
    group_reverse.loc[:, "position"] = -group.loc[:, "position"]
    group_reverse.loc[:, "|x|"] = group.loc[:, "|x|"]
    group_reverse.loc[:, "sample"] = group.loc[:, "sample"]
    return group_reverse


def append_reverse_groups(df_mismatch):
    groups = []
    for i, group in df_mismatch.groupby(["tax_id"], sort=False):
        group_reverse = make_reverse_group(group)
        group_combined = pd.concat([group, group_reverse])
        groups.append(group_combined)
    df_mismatch = pd.concat(groups)
    return df_mismatch


def df_mismatch_to_mapDamage(df_mismatch):

    df_mapDamage = df_mismatch.copy()

    # if forward only, fill in zeros for reverse direction
    if "3'" not in df_mapDamage["direction"].unique():
        df_mapDamage = append_reverse_groups(df_mapDamage)

    bases = ["A", "C", "G", "T"]
    for base in bases:
        if base not in df_mapDamage.columns:
            mismatches.add_reference_count(df_mapDamage, ref=base)

    df_mapDamage["Total"] = df_mapDamage[bases].sum(axis=1)

    # fmt: off
    columns = [
        "Chr", "End", "Std", "Pos",
        "A", "C", "G", "T", "Total",
        "G>A", "C>T", "A>G", "T>C", "A>C", "A>T", "C>G", "C>A", "T>G", "T>A", "G>C", "G>T",
        "A>-", "T>-", "C>-", "G>-",
        "->A", "->T", "->C", "->G",
        "S",
    ]
    # fmt: on

    d_rename = {
        "tax_id": "Chr",
        "position": "Pos",
    }

    idx_start = columns.index("G>A")
    idx_end = columns.index("G>T") + 1
    for sub in columns[idx_start:idx_end]:
        d_rename[sub[0] + sub[-1]] = sub

    df_mapDamage = df_mapDamage.rename(columns=d_rename)
    df_mapDamage["End"] = np.where(df_mapDamage["Pos"] > 0, "5p", "3p")
    df_mapDamage["Std"] = "+"
    df_mapDamage["Pos"] = np.abs(df_mapDamage["Pos"])

    for col in columns[idx_end:]:
        df_mapDamage[col] = 0

    df_mapDamage = df_mapDamage.loc[:, columns]

    return df_mapDamage


def convert(filename, csv_out):

    df_mismatch = pd.read_parquet(filename)
    df_mapDamage = df_mismatch_to_mapDamage(df_mismatch)

    out = ""
    out += (
        f"# table produced by metaDMG version {version} \n"
        f"# using mismatch file {filename.name} \n"
        f"# Chr: Tax ID, "
        f"End: from which termini of DNA sequences, "
        f"Std: Defined to be relative to + strand \n"
    )

    out += df_mapDamage.to_csv(index=False, sep="\t")

    with open(csv_out, "w", encoding="utf-8") as file:
        file.write(out)



#%%
import jax.numpy as jnp
import numpy as np
import numpyro
import pandas as pd
from jax import jit
from jax.random import PRNGKey as Key
from numba import njit
from numpyro import distributions as dist
from numpyro.infer import MCMC, NUTS, Predictive, log_likelihood
from scipy.special import logsumexp
from scipy.stats import beta as sp_beta
from scipy.stats import betabinom as sp_betabinom
from scipy.stats import norm as sp_norm

from metaDMG.fit import fit_utils


#%%

numpyro.enable_x64()

priors = fit_utils.get_priors()
A_prior = priors["A"]  # mean = 0.01, concentration = 1
q_prior = priors["q"]  # mean = 0.2, concentration = 5
c_prior = priors["c"]  # mean = 0.1, concentration = 10
phi_prior = priors["phi"]

#%%


def numpyro_model(x, N, k=None):
    x_abs = jnp.abs(x)

    A = numpyro.sample("A", dist.Beta(A_prior[0], A_prior[1]))
    q = numpyro.sample("q", dist.Beta(q_prior[0], q_prior[1]))
    c = numpyro.sample("c", dist.Beta(c_prior[0], c_prior[1]))
    Dx = jnp.clip(numpyro.deterministic("Dx", A * (1 - q) ** (x_abs - 1) + c), 0, 1)

    delta = numpyro.sample("delta", dist.Exponential(1 / phi_prior[1]))
    phi = numpyro.deterministic("phi", delta + phi_prior[0])

    alpha = numpyro.deterministic("alpha", Dx * phi)
    beta = numpyro.deterministic("beta", (1 - Dx) * phi)

    numpyro.sample("obs", dist.BetaBinomial(alpha, beta, N), obs=k)


#%%


def filter_out_k(data):
    return {key: value for key, value in data.items() if key != "k"}


@jit
def _get_posterior(rng_key, samples, *args, **kwargs):
    return Predictive(numpyro_model, samples)(rng_key, *args, **kwargs)


def get_posterior_predictive(mcmc, data):
    posterior_samples = mcmc.get_samples()
    rng_key = Key(0)
    data_no_k = filter_out_k(data)
    return _get_posterior(rng_key, posterior_samples, **data_no_k)


def get_posterior_predictive_obs(mcmc, data):
    return get_posterior_predictive(mcmc, data)["obs"]


#%%


def get_n_sigma_probability(n_sigma):
    return sp_norm.cdf(n_sigma) - sp_norm.cdf(-n_sigma)


CONF_1_SIGMA = get_n_sigma_probability(1)


def add_D_information(
    fit_result,
    mcmc,
    data,
    prefix="",
):

    A = mcmc.get_samples()["A"]
    phi = mcmc.get_samples()["phi"]
    N = max(data["N"][0], 1)
    mu = np.mean(A)
    std = np.mean(np.sqrt(A * (1 - A) * (phi + N) / ((phi + 1) * N)))

    Dx = A
    alpha = Dx * phi
    beta = (1 - Dx) * phi

    # 1000x faster approximation for sp_betabinom(N, alpha, beta)
    pdf_approx = sp_betabinom(N, alpha.mean(), beta.mean())

    fit_result[f"{prefix}damage"] = mu.item()
    fit_result[f"{prefix}damage_std"] = std.item()
    fit_result[f"{prefix}significance"] = mu.item() / std.item()
    fit_result[f"{prefix}damage_median"] = np.median(A)

    # fit_result[f"{prefix}prob_lt_1p_damage"] = pdf_beta.cdf(0.01).mean()
    # # fit_result[f"{prefix}prob_lt_1p_damage_betabinom"] = pdf.cdf(0.01 * N).mean()
    # fit_result[f"{prefix}prob_zero_damage"] = pdf.cdf(0).mean()

    for n_sigma in [1, 2, 3]:
        conf = get_n_sigma_probability(n_sigma)
        fit_result[f"{prefix}damage_CI_{n_sigma}_sigma_low"] = (
            pdf_approx.ppf((1 - conf) / 2.0).mean() / N
        )
        fit_result[f"{prefix}damage_CI_{n_sigma}_sigma_high"] = (
            pdf_approx.ppf((1 + conf) / 2.0).mean() / N
        )

    fit_result[f"{prefix}damage_CI_95_low"] = (
        pdf_approx.ppf((1 - 0.95) / 2.0).mean() / N
    )
    fit_result[f"{prefix}damage_CI_95_high"] = (
        pdf_approx.ppf((1 + 0.95) / 2.0).mean() / N
    )

    return fit_result


#%%


@jit
def _compute_log_likelihood(posterior_samples, data):
    return log_likelihood(numpyro_model, posterior_samples, **data)["obs"]


def compute_log_likelihood(mcmc, data):
    posterior_samples = mcmc.get_samples()
    return _compute_log_likelihood(posterior_samples, data)


#%%


def add_summary_of_variable(
    fit_result,
    mcmc,
    variable,
    prefix="",
    axis=0,
):
    values = np.array(mcmc.get_samples()[variable])

    s = f"{prefix}{variable}"
    q_low = (1 - CONF_1_SIGMA) / 2.0
    q_high = (1 + CONF_1_SIGMA) / 2.0

    fit_result[f"{s}"] = np.mean(values, axis=axis)
    fit_result[f"{s}_std"] = np.std(values, axis=axis)
    fit_result[f"{s}_median"] = np.median(values, axis=axis)
    fit_result[f"{s}_CI_1_sigma_low"] = np.quantile(values, q_low, axis=axis)
    fit_result[f"{s}_CI_1_sigma_high"] = np.quantile(values, q_high, axis=axis)


def compute_rho_Ac(mcmc):
    A = mcmc.get_samples()["A"]
    c = mcmc.get_samples()["c"]
    return np.corrcoef(A, c)[0, 1]


#%%


def _init_mcmc(model, **kwargs):

    mcmc_kwargs = dict(
        progress_bar=False,
        num_warmup=500,
        num_samples=1000,
        num_chains=1,  # problem when setting to 2
        chain_method="sequential",
        # http://num.pyro.ai/en/stable/_modules/numpyro/infer/mcmc.html#MCMC
    )

    return MCMC(NUTS(model), jit_model_args=True, **mcmc_kwargs, **kwargs)


def init_mcmc(config, **kwargs):
    if config["bayesian"]:
        mcmc = _init_mcmc(numpyro_model, **kwargs)
    else:
        mcmc = None
    return mcmc


def fit_mcmc(mcmc, data, seed=0):
    mcmc.run(Key(seed), **data)


def use_last_state_as_warmup_state(mcmc):
    # https://github.com/pyro-ppl/numpyro/issues/539
    mcmc._warmup_state = mcmc._last_state


def add_Bayesian_fit_result(
    fit_result,
    data,
    mcmc,
):

    add_D_information(fit_result, mcmc, data)

    add_summary_of_variable(fit_result, mcmc, "A")
    add_summary_of_variable(fit_result, mcmc, "q")
    add_summary_of_variable(fit_result, mcmc, "c")
    add_summary_of_variable(fit_result, mcmc, "phi")

    fit_result["rho_Ac"] = compute_rho_Ac(mcmc)


def make_fits(
    fit_result,
    data,
    mcmc,
):
    fit_mcmc(mcmc, data)
    add_Bayesian_fit_result(
        fit_result,
        data,
        mcmc,
    )

    # mcmc.print_summary(prob=0.68)
    # if False:
    #     use_last_state_as_warmup_state(mcmc)



import shutil
from importlib import resources
from pathlib import Path


def get_data(output_dir: Path = Path("raw_data")) -> None:
    """Copies the data file to a new directory

    Parameters
    ----------
    output_dir
        _description_
    """

    files = [
        "alignment.sorted.bam",
        "names-mdmg.dmp",
        "nodes-mdmg.dmp",
        "acc2taxid.map.gz",
    ]

    for file in files:
        with resources.path("metaDMG.data", file) as p:
            file_path = p

        target_path = Path(output_dir) / file
        target_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(file_path, target_path)



import itertools

import dash_bootstrap_components as dbc
import numpy as np
from dash import dcc, html

from metaDMG.viz import figures, viz_utils


#%%


def get_navbar():

    navbar = dbc.NavbarSimple(
        [
            dbc.Button(
                "Filters",
                outline=True,
                color="light",
                className="mr-1",
                id="sidebar_left_toggle_btn",
            ),
            dbc.Button(
                "Styling",
                outline=True,
                color="light",
                className="mr-1",
                id="navbar_btn_toggle_styling",
            ),
            dbc.Button(
                "Raw data",
                outline=False,
                color="light",
                className="mr-1",
                id="sidebar_right_toggle_btn",
            ),
            dbc.Button(
                "",
                color="link",
                outline=False,
                disabled=True,
                id="invisible_button",
            ),
            html.Div(dcc.Store(id="memory-store-csv")),
            html.Div(dcc.Store(id="memory-store-pdf")),
            dbc.Button(
                "Export CSV",
                outline=True,
                color="light",
                className="mr-1",
                id="navbar_btn_export_csv",
            ),
            dcc.Download(id="export_csv"),
            dbc.Button(
                "Export PDF",
                outline=True,
                color="light",
                className="mr-1",
                id="navbar_btn_export_pdf",
            ),
            dcc.Download(id="export_pdf"),
        ],
        brand="metaDMG",
        brand_href="https://github.com/metaDMG/metaDMG",
        color="dark",
        dark=True,
        fluid=True,
    )

    return navbar


#%%


def get_content_main(viz_results, start_configuration):

    _, columns, columns_no_log = viz_utils.get_d_columns_latex(viz_results)

    dropdown_x_axis = dcc.Dropdown(
        id="xaxis_column",
        options=[{"label": i, "value": i} for i in columns],
        value="significance" if viz_results.Bayesian else "MAP_significance",
    )

    dropdown_y_axis = dcc.Dropdown(
        id="yaxis_column",
        options=[{"label": i, "value": i} for i in columns],
        value="damage" if viz_results.Bayesian else "MAP_damage",
    )

    XY_axis_dropdowns = [
        dbc.Col(
            [
                dbc.Row(
                    [
                        dbc.Col(html.Center("X-axis: ")),
                        dbc.Col(dropdown_x_axis, width=12),
                    ]
                ),
            ],
            width=6,
        ),
        dbc.Col(
            [
                dbc.Row(
                    [
                        dbc.Col(html.Center("Y-axis: ")),
                        dbc.Col(dropdown_y_axis, width=12),
                    ]
                ),
            ],
            width=6,
        ),
    ]

    marker_transformation_function = dcc.Dropdown(
        id="marker_transformation_function",
        options=[
            {"label": "Constant", "value": "constant"},
            {"label": "Linear", "value": "linear"},
            {"label": "Sqrt", "value": "sqrt"},
            {"label": "Log", "value": "log10"},
        ],
        value="sqrt",
        searchable=False,
        clearable=False,
    )

    marker_transformation_variable = dcc.Dropdown(
        id="marker_transformation_variable",
        options=[{"label": col, "value": col} for col in columns_no_log],
        value="N_reads",
        searchable=True,
        clearable=False,
    )

    marker_transformation_slider = dcc.Slider(
        id="marker_transformation_slider",
        min=1,
        max=60,
        step=1,
        value=30,
        marks={mark: str(mark) for mark in [1, 10, 20, 30, 40, 50, 60]},
    )

    marker_transformations = [
        dbc.Col(
            dbc.Row(
                [
                    dbc.Col(html.Center("Variable:")),
                    dbc.Col(marker_transformation_variable, width=12),
                ],
            ),
            width=4,
        ),
        dbc.Col(
            dbc.Row(
                [
                    dbc.Col(html.Center("Function:")),
                    dbc.Col(marker_transformation_function, width=12),
                ],
            ),
            width=2,
        ),
        dbc.Col(
            dbc.Row(
                [
                    dbc.Col(html.Center("Scale:")),
                    dbc.Col(marker_transformation_slider, width=12),
                ],
            ),
            width=6,
        ),
    ]

    content_main = html.Div(
        html.Div(
            [
                dcc.Graph(id="main_graph", **viz_utils.get_graph_kwargs()),
                dbc.Collapse(
                    [
                        dbc.Row(dbc.Col(html.Hr())),
                        dbc.Row(
                            dbc.Col(html.Center("Axis variables", className="lead"))
                        ),
                        dbc.Row(
                            XY_axis_dropdowns,
                            # no_gutters=True,
                            className="g-0",
                        ),
                        dbc.Row(dbc.Col(html.Hr())),
                        dbc.Row(dbc.Col(html.Center("Marker size", className="lead"))),
                        dbc.Row(
                            marker_transformations,
                            # no_gutters=True,
                            className="g-0",
                        ),
                    ],
                    id="navbar_collapsed_toggle_styling",
                    is_open=False,
                ),
            ]
        ),
        id="content_main",
        style=start_configuration.style_content_main,
    )

    return content_main


#%%


#%%


def get_sidebar_left(viz_results, start_configuration):

    d_columns_latex, columns, columns_no_log = viz_utils.get_d_columns_latex(
        viz_results
    )

    filter_dropdown_file = dbc.Form(
        viz_utils.get_dropdown_file_selection(
            viz_results=viz_results,
            id="sidebar_left_dropdown_samples",
            samples_to_show="each",  # one for each first letter in sample
        ),
    )

    filters_collapse_files = html.Div(
        [
            html.Div(
                dbc.Button(
                    "Samples",
                    id="sidebar_left_samples_btn",
                    color="secondary",
                    outline=True,
                    size="lg",
                ),
                className="d-grid gap-2",
            ),
            dbc.Collapse(
                filter_dropdown_file,
                id="sidebar_left_samples_collapsed",
                is_open=False,
            ),
        ]
    )

    filter_tax_id = dbc.Row(
        [
            dbc.Col(html.Br(), width=12),
            dbc.Col(html.H6("Specific taxas:"), width=12),
            dbc.Col(
                dbc.Form(
                    dcc.Dropdown(
                        id="sidebar_left_taxa_input_specific",
                        options=[
                            {"label": tax, "value": tax}
                            for tax in itertools.chain.from_iterable(
                                [
                                    viz_results.all_tax_ranks,
                                    viz_results.all_tax_names,
                                    viz_results.all_tax_ids,
                                ]
                            )
                        ],
                        clearable=True,
                        multi=True,
                        placeholder="Select taxas...",
                    ),
                ),
                width=12,
            ),
            dbc.Col(html.Br(), width=12),
            dbc.Col(html.H6("Taxonomic path contains:"), width=12),
            dbc.Col(
                dbc.Form(
                    dbc.Input(
                        id="sidebar_left_taxa_input_path_contains",
                        placeholder="Input goes here...",
                        type="text",
                        autoComplete="off",
                        debounce=True,  # activates after enter
                    ),
                ),
                width=12,
            ),
            dbc.Col(
                dbc.Form(
                    html.Div(
                        dbc.Button(
                            "Update",
                            id="tax_id_plot_button",
                            color="light",
                        ),
                        className="d-grid gap-2",
                    ),
                ),
            ),
        ],
        justify="between",
        className="form-row"
        # form=True,
    )

    filters_collapse_tax_id = html.Div(
        [
            html.Div(
                dbc.Button(
                    "Taxa",
                    id="sidebar_left_taxanomics_btn",
                    color="secondary",
                    outline=True,
                    size="lg",
                ),
                className="d-grid gap-2",
            ),
            dbc.Collapse(
                filter_tax_id,
                id="sidebar_left_taxanomics_collapsed",
                is_open=False,
            ),
        ]
    )

    #%%

    # default sliders (with default values)
    slider_N_reads = make_new_slider(
        viz_results,
        column="N_reads",
        id_type="dbc",
        min_value=np.log10(100),
    )
    slider_k_sum_total = make_new_slider(
        viz_results,
        column="k_sum_total",
        id_type="dbc",
        min_value=np.log10(10),
    )

    slider_phi_column = "phi" if viz_results.Bayesian else "MAP_phi"
    slider_phi = make_new_slider(
        viz_results,
        column=slider_phi_column,
        id_type="dbc",
        min_value=np.log10(100),
    )

    filters_collapse_ranges = html.Div(
        [
            html.Div(
                dbc.Button(
                    "Fits",
                    id="sidebar_left_results_btn",
                    color="secondary",
                    outline=True,
                    size="lg",
                ),
                className="d-grid gap-2",
            ),
            dbc.Collapse(
                [
                    html.Br(),
                    dbc.Col(
                        html.H6("Fit results:"),
                        width=12,
                    ),
                    dbc.Col(
                        dcc.Dropdown(
                            id="sidebar_left_results",
                            options=[
                                {"label": sample, "value": sample}
                                for sample in columns_no_log
                            ],
                            value=[
                                "N_reads",
                                "k_sum_total",
                                slider_phi_column,
                            ],
                            multi=True,
                            placeholder="Select a variable ...",
                            optionHeight=30,
                        ),
                        width=12,
                    ),
                    dbc.Col(
                        id="sidebar_left_results_container",
                        # children=[],
                        children=[
                            slider_N_reads,
                            slider_k_sum_total,
                            slider_phi,
                        ],
                        width=12,
                    ),
                    html.Br(),
                    html.Br(),
                    html.Br(),
                ],
                id="sidebar_left_results_collapsed",
                is_open=False,
            ),
        ]
    )

    sidebar_left = html.Div(
        [
            html.H2("Filters", className="display-4"),
            dbc.Form(
                [
                    filters_collapse_files,
                    html.Hr(),
                    filters_collapse_tax_id,
                    html.Hr(),
                    filters_collapse_ranges,
                ]
            ),
        ],
        id="sidebar_left",
        style=start_configuration.style_sidebar_left,
    )

    return sidebar_left


#%%


def get_sidebar_right(start_configuration):

    sidebar_right_results = html.Div(
        [
            html.Div(
                dbc.Button(
                    "Fit results",
                    id="sidebar_right_btn_toggle_results",
                    color="secondary",
                    outline=True,
                    size="lg",
                ),
                className="d-grid gap-2",
            ),
            dbc.Collapse(
                html.Div(
                    "sidebar_right_datatable_results",
                    id="sidebar_right_datatable_results",
                ),
                id="sidebar_right_collapsed_toggle_results",
                is_open=False,
            ),
        ]
    )

    sidebar_right_collapsed_toggle_combined_graph = dbc.Form(
        dcc.Graph(
            figure=figures.create_empty_figure(),
            id="sidebar_right_graph_combined",
            **viz_utils.get_graph_kwargs_no_buttons(),
        ),
    )

    sidebar_right_collapsed_toggle_combined = html.Div(
        [
            html.Div(
                dbc.Button(
                    "Combined",
                    id="sidebar_right_btn_toggle_combined",
                    color="secondary",
                    outline=False,
                    size="lg",
                ),
                className="d-grid gap-2",
            ),
            dbc.Collapse(
                sidebar_right_collapsed_toggle_combined_graph,
                id="sidebar_right_collapsed_toggle_combined",
                is_open=True,
            ),
        ]
    )

    # sidebar_right_collapsed_toggle_forward_only_graph = dbc.Form(
    #     [
    #         dcc.Graph(
    #             figure=figures.create_empty_figure(),
    #             id="sidebar_right_graph_forward",
    #             style={"height": "20vh"},
    #             **viz_utils.get_graph_kwargs_no_buttons(),
    #         ),
    #         # dcc.Graph(
    #         #     figure=figures.create_empty_figure(),
    #         #     id="sidebar_right_graph_reverse",
    #         #     style={"height": "20vh"},
    #         #     **viz_utils.get_graph_kwargs_no_buttons(),
    #         # ),
    #     ]
    # )

    # sidebar_right_collapsed_toggle_forward_only = html.Div(
    #     [
    #         html.Div(
    #             dbc.Button(
    #                 "Forward / Reverse",
    #                 id="sidebar_right_btn_toggle_forward_only",
    #                 color="secondary",
    #                 outline=True,
    #                 size="lg",
    #             ),
    #             className="d-grid gap-2",
    #         ),
    #         dbc.Collapse(
    #             sidebar_right_collapsed_toggle_forward_only_graph,
    #             id="sidebar_right_collapsed_toggle_forward_only",
    #             is_open=False,
    #         ),
    #         html.Br(),
    #         html.Br(),
    #         html.Br(),
    #     ]
    # )

    sidebar_right = html.Div(
        [
            html.H2("Raw data", className="display-4"),
            # html.Hr(),
            sidebar_right_collapsed_toggle_combined,
            html.Hr(),
            sidebar_right_results,
            # html.Hr(),
            # sidebar_right_collapsed_toggle_forward_only,
        ],
        id="sidebar_right",
        style=start_configuration.style_sidebar_right,
    )

    return sidebar_right


#%%


def get_app_layout(viz_results, start_configuration):

    navbar = get_navbar()
    content_main = get_content_main(viz_results, start_configuration)
    sidebar_left = get_sidebar_left(viz_results, start_configuration)
    sidebar_right = get_sidebar_right(start_configuration)

    modal_filtering = dbc.Modal(
        [
            dbc.ModalHeader("Filtering Error"),
            dbc.ModalBody(
                "Too restrictive filtering, no points left to plot. "
                "Please choose a less restrictive filtering."
            ),
            dbc.ModalFooter(
                dbc.Button("Close", id="modal_close_button", className="ml-auto")
            ),
        ],
        centered=True,
        id="modal",
    )

    modal_pdf = dbc.Modal(
        [
            dbc.ModalHeader("Generating plots, please wait."),
            dbc.ModalBody(
                dbc.Progress(
                    id="progress_bar",
                    value=5,
                    max=100,
                    color="info",
                    striped=True,
                    animated=True,
                )
            ),
            dbc.ModalFooter(
                dbc.Button(
                    "Cancel",
                    id="modal_pdf_close_button",
                    className="ml-auto",
                ),
            ),
        ],
        centered=True,
        id="modal_pdf",
    )

    return html.Div(
        [
            dcc.Store(id="sidebar_right_state"),
            dcc.Store(id="sidebar_left_state"),
            navbar,
            sidebar_left,
            content_main,
            sidebar_right,
            modal_filtering,
            modal_pdf,
        ],
    )


#%%


def get_id_dict(child):
    return child["props"]["id"]


def find_index_in_children(children, id_type, search_index):
    for i, child in enumerate(children):
        d_id = get_id_dict(child)
        if d_id["type"] == id_type and d_id["index"] == search_index:
            return i


def get_current_names(current_ids):
    return [x["index"] for x in current_ids if x]


def slider_is_added(current_names, dropdown_names):
    "Returns True if a new slider is added, False otherwise"
    return set(current_names).issubset(dropdown_names)


def get_name_of_added_slider(current_names, dropdown_names):
    return list(set(dropdown_names).difference(current_names))[0]


def get_name_of_removed_slider(current_names, dropdown_names):
    return list(set(current_names).difference(dropdown_names))


def remove_name_from_children(column, children, id_type):
    "Given a column, remove the corresponding child element from children"
    index = find_index_in_children(children, id_type=id_type, search_index=column)
    children.pop(index)


def get_slider_name(column, low_high):
    if isinstance(low_high, dict):
        low = low_high["min"]
        high = low_high["max"]
    elif isinstance(low_high, (tuple, list)):
        low = low_high[0]
        high = low_high[1]

    if viz_utils.is_log_transform_column(column):
        low = viz_utils.log_transform_slider(low)
        high = viz_utils.log_transform_slider(high)

    low = viz_utils.human_format(low)
    high = viz_utils.human_format(high)

    return f"{column}: [{low}, {high}]"


def make_new_slider(
    viz_results,
    column,
    id_type,
    N_steps=100,
    value=None,
    min_value=None,
):

    d_range_slider = viz_utils.get_range_slider_keywords(
        viz_results,
        column=column,
        N_steps=N_steps,
    )

    if value is not None:
        d_range_slider["value"] = value

    if min_value is not None:
        d_range_slider["value"][0] = min_value

    return dbc.Container(
        [
            dbc.Row(html.Br()),
            dbc.Row(
                html.P(
                    get_slider_name(column, d_range_slider),
                    id={
                        "type": "sidebar_left_results_dynamic_name",
                        "index": column,
                    },
                ),
                justify="center",
            ),
            dbc.Row(
                dbc.Col(
                    dcc.RangeSlider(
                        id={
                            "type": "sidebar_left_results_dynamic",
                            "index": column,
                        },
                        **d_range_slider,
                    ),
                    width=12,
                ),
            ),
        ],
        id={"type": id_type, "index": column},
    )
